/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 23/07/22.
//

#ifndef FGGL_ENTITY_LOADER_LOADER_HPP
#define FGGL_ENTITY_LOADER_LOADER_HPP

#include <functional>
#include <map>
#include <utility>

#include "fggl/util/guid.hpp"
#include "fggl/modules/module.hpp"
#include "fggl/entity/entity.hpp"
#include "fggl/entity/loader/spec.hpp"
#include "fggl/assets/loader.hpp"

namespace fggl::entity {

	constexpr auto PROTOTYPE_ASSET = assets::AssetType::make("entity_prototype");
	constexpr auto SCENE = assets::AssetType::make("entity_scene");

	using FactoryFunc = std::function<void(const ComponentSpec &config, EntityManager &, const EntityID &, modules::Services &svc)>;
	using CustomiseFunc = std::function<void(EntityManager &, const EntityID &)>;

	struct FactoryInfo {
		FactoryFunc factory;
		CustomiseFunc finalise = nullptr;
	};

	class EntityFactory {
		public:
			constexpr static const modules::ModuleService service = modules::make_service("fggl::entity:Factory");

			inline EntityFactory(modules::Services &services) : m_services(services) {}

			EntityID create(const EntityType &spec, EntityManager &manager, const CustomiseFunc &customise = nullptr) {
				std::vector<CustomiseFunc> finishers;

				// build the setup
				auto entity = setupComponents(spec, manager, finishers);
				if ( entity == entity::INVALID ) {
					debug::error("EntityFactory: failed to build from prototype {}", std::to_string(spec.get()));
					return entity::INVALID;
				}

				// if requested, allow the user to customize the creation
				if ( customise != nullptr ) {
					customise(manager, entity);
				}

				// run finishers for components
				for ( auto& finisher : finishers ) {
					finisher( manager, entity );
				}

				// use metadata to finalise
				processTags(manager, entity, spec);

				return entity;
			}

			void processTags(EntityManager& manager, EntityID id, EntityType spec) {
				auto type = m_prototypes.at(spec);
				for ( auto& tag : type.tags ) {
					manager.addTag(id, tag);
				}
			}

			void log_known_types() const {
				debug::debug("dumping known types:");
				for(const auto& [k,v] : m_factories) {
					debug::debug("\ttype: {}", k);
				}
			}

			void define(EntityType type, const EntitySpec &spec) {
				m_prototypes[type] = spec;
			}

			// ability to set and unset factory functions
			inline void bind(const ComponentID &configNode, FactoryFunc factory, CustomiseFunc finalise = nullptr) {
				m_factories[configNode].factory = std::move(factory);
				m_factories[configNode].finalise = std::move(finalise);
			}

			inline void unbind(const ComponentID &configNode) {
				m_factories.erase(configNode);
			}

			inline FactoryInfo& getInfo(ComponentID comp) {
				return m_factories.at(comp);
			}

		private:
			modules::Services m_services;
			std::map<ComponentID, FactoryInfo> m_factories;
			std::map<EntityType, EntitySpec> m_prototypes;

			entity::EntityID setupComponents(EntityType entityType,
											 EntityManager &manager,
											 std::vector<CustomiseFunc> &finishers) {
				assert(entityType != NO_PARENT && "setup components called with NO_PARENT?!");

				auto entity = manager.create();
				std::vector<ComponentID> loadedComps;

				auto currentType = entityType;
				while (currentType != NO_PARENT) {
					const auto& specEntry = m_prototypes.find( currentType );
					if ( specEntry == m_prototypes.end() ) {
						debug::warning("Asked to setup {}, for {} but was not a known prototype", specEntry->first, entityType);
						return entity::INVALID;
					}

					auto entitySpec = specEntry->second;
					debug::debug("constructing {} for {} ({} comps)", currentType, entityType, entitySpec.components.size());
					assert( entitySpec.ordering.size() == entitySpec.components.size() && "ordering incorrect size, bad things happend!" );

					for (auto &component : entitySpec.ordering) {
						// skip comps loaded by children
						if (std::find(loadedComps.begin(), loadedComps.end(), component) != loadedComps.end()) {
							continue;
						}

						try {
							auto &data = getComponent(entitySpec, component);
							loadedComps.push_back(component);

							auto &info = m_factories.at(component);
							info.factory(data, manager, entity, m_services);

							if (info.finalise != nullptr) {
								finishers.push_back(info.finalise);
							}
						} catch (std::out_of_range &ex) {
							debug::error( "EntityFactory: Unknown component factory type '{}'", component );
							log_known_types();
							manager.destroy(entity);
							return entity::INVALID;
						}
					}
					currentType = entitySpec.parent;
				}

				return entity;
			}

			ComponentSpec &getComponent(EntitySpec &prototype, util::GUID compToken) {
				auto compItr = prototype.components.find(compToken);
				if (compItr != prototype.components.end()) {
					return compItr->second;
				}

				if (prototype.parent == NO_PARENT) {
					throw std::out_of_range("EntityFactory: no such component!");
				}
				return getComponent(m_prototypes.at(prototype.parent), compToken);
			}
	};

	assets::AssetRefRaw load_prototype(assets::Loader* loader, EntityFactory *factory, const assets::AssetGUID &guid, assets::AssetData data);
	assets::AssetRefRaw load_scene(assets::Loader* loader, const assets::AssetGUID& asset, assets::AssetData data, void* ptr);

} // namespace fggl::entity

#endif //FGGL_ENTITY_LOADER_LOADER_HPP
